/** 
 * 
 * Copyright 2004 Protique Ltd
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License. 
 * 
 **/
package org.codehaus.activemq.broker;

import org.codehaus.activemq.capacity.CapacityMonitor;
import org.codehaus.activemq.message.ActiveMQMessage;
import org.codehaus.activemq.message.ActiveMQXid;
import org.codehaus.activemq.message.ConsumerInfo;
import org.codehaus.activemq.message.MessageAck;
import org.codehaus.activemq.service.Service;
import org.codehaus.activemq.store.PersistenceAdapter;

import javax.jms.JMSException;
import javax.transaction.xa.XAException;
import java.io.File;

/**
 * The Message Broker which routes messages,
 * maintains subscriptions and connections, acknowlegdges messages and handles
 * transactions.
 *
 * @version $Revision: 1.7 $
 */
public interface Broker extends Service, CapacityMonitor {

    /**
     * Add an active message consumer
     *
     * @param client
     * @param info
     * @throws JMSException
     */
    public void addMessageConsumer(BrokerClient client, ConsumerInfo info) throws JMSException;

    /**
     * remove an active message consumer
     *
     * @param client
     * @param info
     * @throws JMSException
     */
    public void removeMessageConsumer(BrokerClient client, ConsumerInfo info) throws JMSException;

    /**
     * gets a list of all the prepared xa transactions.
     *
     * @param client
     */
    public ActiveMQXid[] getPreparedTransactions(BrokerClient client) throws XAException;

    /**
     * Acknowledge consumption of a message by the Message Consumer
     *
     * @param client
     * @param ack
     * @throws JMSException
     */
    public void acknowledgeMessage(BrokerClient client, MessageAck ack) throws JMSException;

    /**
     * Acknowledge consumption of a message within a transaction
     *
     * @param client
     * @param transactionId
     * @param ack
     * @throws JMSException
     */
    public void acknowledgeTransactedMessage(BrokerClient client, String transactionId, MessageAck ack) throws JMSException;


    /**
     * Called after a rollback of a JMS transaction to redeliver the message to the consumers
     * dispatch queue
     *
     * @param client
     * @param ack
     */
    public void redeliverMessage(BrokerClient client, MessageAck ack) throws JMSException;

    /**
     * send a message to the broker
     *
     * @param client
     * @param message
     * @throws JMSException
     */
    public void sendMessage(BrokerClient client, ActiveMQMessage message) throws JMSException;

    /**
     * send a message to the broker within a transaction
     *
     * @param client
     * @param transactionId
     * @param message
     * @throws JMSException
     */
    public void sendTransactedMessage(BrokerClient client, String transactionId, ActiveMQMessage message) throws JMSException;

    /**
     * A hint to the broker that an BrokerClient has stopped
     * This enables the broker to clean-up any outstanding processing
     * that may be outstanding
     *
     * @param client
     * @throws JMSException
     */
    public void cleanUpClient(BrokerClient client) throws JMSException;

    /**
     * Delete a durable subscriber
     *
     * @param clientId
     * @param subscriberName
     * @throws JMSException if the subscriber doesn't exist or is still active
     */
    public void deleteSubscription(String clientId, String subscriberName) throws JMSException;

    /**
     * start a transaction
     *
     * @param client
     * @param transactionId
     */
    public void startTransaction(BrokerClient client, String transactionId) throws JMSException;

    /**
     * commit a transaction
     *
     * @param client
     * @param transactionId
     * @throws JMSException
     */
    public void commitTransaction(BrokerClient client, String transactionId) throws JMSException;

    /**
     * rollback a transaction
     *
     * @param client
     * @param transactionId
     * @throws JMSException
     */
    public void rollbackTransaction(BrokerClient client, String transactionId) throws JMSException;


    /**
     * @param client
     * @param xid
     */
    public void startTransaction(BrokerClient client, ActiveMQXid xid) throws XAException;

    /**
     * @param client
     * @param xid
     */
    public int prepareTransaction(BrokerClient client, ActiveMQXid xid) throws XAException;

    /**
     * @param client
     * @param xid
     */
    public void rollbackTransaction(BrokerClient client, ActiveMQXid xid) throws XAException;

    /**
     * @param client
     * @param xid
     * @param onePhase
     */
    public void commitTransaction(BrokerClient client, ActiveMQXid xid, boolean onePhase) throws XAException;

    /**
     * Get a temp directory - used for spooling
     *
     * @return a File ptr to the directory
     */
    public File getTempDir();

    public String getBrokerName();

    PersistenceAdapter getPersistenceAdapter();

    void setPersistenceAdapter(PersistenceAdapter persistenceAdapter);
}
