/** 
 * 
 * Copyright 2004 Protique Ltd
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License. 
 * 
 **/

package org.codehaus.activemq.message;

/**
 * Abstract class for a transportable Packet
 *
 * @version $Revision: 1.10 $
 */
public abstract class AbstractPacket implements Packet {

    private String id;
    private boolean receiptRequired;
    private transient int memoryUsage = 2048;
    private transient int memoryUsageReferenceCount;

    /**
     * @return the unique id for this Packet
     */

    public String getId() {
        return this.id;
    }


    /**
     * Set the unique id for this Packet
     *
     * @param newId
     */

    public void setId(String newId) {
        this.id = newId;
    }


    /**
     * @return true if a Recipt is required
     */

    public boolean isReceiptRequired() {
        return this.receiptRequired;
    }

    /**
     * @return false since most packets are not receipt packets
     */
    public boolean isReceipt() {
        return false;
    }

    /**
     * Set if a Recipt if required on receiving this Packet
     *
     * @param value
     */

    public void setReceiptRequired(boolean value) {
        this.receiptRequired = value;
    }

    /**
     * Retrieve if a JMS Message type or not
     *
     * @return true if it is a JMS Message
     */
    public boolean isJMSMessage() {
        return false;
    }

    /**
     * Tests equality with another instance
     *
     * @param obj - the other instance to test equality with
     * @return Returns true if the objects are equilvant
     */

    public boolean equals(Object obj) {
        boolean result = this == obj;
        if (!result && obj != null && obj instanceof AbstractPacket) {
            AbstractPacket other = (AbstractPacket) obj;
            result = other.getId().equals(this.getId());
        }

        return result;
    }

    /**
     * @return Returns hash code for this instance
     */

    public int hashCode() {
        return this.id != null ? this.id.hashCode() : super.hashCode();
    }
    
    /**
     * Get a hint about how much memory this Packet is consuming
     * @return an aproximation of the current memory used by this instance
     */
    public int getMemoryUsage(){
        return memoryUsage;
    }
    
    /**
     * Set a hint about how mujch memory this packet is consuming
     * @param newMemoryUsage
     */
    public void setMemoryUsage(int newMemoryUsage){
        this.memoryUsage = newMemoryUsage;
    }
    
    /**
     * Increment reference count for bounded memory collections
     * @return the incremented reference value
     * @see org.codehaus.activemq.message.util.MemoryBoundedQueue
     */
    public synchronized int incrementMemoryReferenceCount(){
        return ++memoryUsageReferenceCount;
    }
    
    
    /**
     * Decrement reference count for bounded memory collections
     * @return the decremented reference value
     * @see org.codehaus.activemq.message.util.MemoryBoundedQueue
     */
    public synchronized int decrementMemoryReferenceCount(){
        return --memoryUsageReferenceCount;
    }
    
    
    /**
     * @return the current reference count for bounded memory collections
     * @see org.codehaus.activemq.message.util.MemoryBoundedQueue
     */
    public synchronized int getMemoryUsageReferenceCount(){
        return memoryUsageReferenceCount;
    }

    /**
     * @return pretty print of this Packet
     */
    public String toString() {
        return getPacketTypeAsString(getPacketType()) + ": " + getId();
    }

    protected static String getPacketTypeAsString(int type) {
        String packetTypeStr = "";
        switch (type) {
            case ACTIVEMQ_MESSAGE:
                packetTypeStr = "ACTIVEMQ_MESSAGE";
                break;
            case ACTIVEMQ_TEXT_MESSAGE:
                packetTypeStr = "ACTIVEMQ_TEXT_MESSAGE";
                break;
            case ACTIVEMQ_OBJECT_MESSAGE:
                packetTypeStr = "ACTIVEMQ_OBJECT_MESSAGE";
                break;
            case ACTIVEMQ_BYTES_MESSAGE:
                packetTypeStr = "ACTIVEMQ_BYTES_MESSAGE";
                break;
            case ACTIVEMQ_STREAM_MESSAGE:
                packetTypeStr = "ACTIVEMQ_STREAM_MESSAGE";
                break;
            case ACTIVEMQ_MAP_MESSAGE:
                packetTypeStr = "ACTIVEMQ_MAP_MESSAGE";
                break;
            case ACTIVEMQ_MSG_ACK:
                packetTypeStr = "ACTIVEMQ_MSG_ACK";
                break;
            case RECEIPT_INFO:
                packetTypeStr = "RECEIPT_INFO";
                break;
            case CONSUMER_INFO:
                packetTypeStr = "CONSUMER_INFO";
                break;
            case PRODUCER_INFO:
                packetTypeStr = "PRODUCER_INFO";
                break;
            case TRANSACTION_INFO:
                packetTypeStr = "TRANSACTION_INFO";
                break;
            case XA_TRANSACTION_INFO:
                packetTypeStr = "XA_TRANSACTION_INFO";
                break;
            case ACTIVEMQ_BROKER_INFO:
                packetTypeStr = "ACTIVEMQ_BROKER_INFO";
                break;
            case ACTIVEMQ_CONNECTION_INFO:
                packetTypeStr = "ACTIVEMQ_CONNECTION_INFO";
                break;
            case SESSION_INFO:
                packetTypeStr = "SESSION_INFO";
                break;
            default:
                packetTypeStr = "UNKNOWN PACKET TYPE: " + type;
        }
        return packetTypeStr;
    }


    /**
     * A helper method used when implementing equals() which returns
     * true if the objects are identical or equal handling nulls properly
     *
     * @return true if the objects are the same or equal or both null
     */
    protected boolean equals(Object left, Object right) {
        return left == right || (left != null && left.equals(right));
    }
}
