/** 
 * 
 * Copyright 2004 Protique Ltd
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License. 
 * 
 **/

package org.codehaus.activemq.service.boundedvm;

import EDU.oswego.cs.dl.util.concurrent.ConcurrentHashMap;
import EDU.oswego.cs.dl.util.concurrent.SynchronizedBoolean;
import org.codehaus.activemq.broker.BrokerClient;
import org.codehaus.activemq.filter.AndFilter;
import org.codehaus.activemq.filter.Filter;
import org.codehaus.activemq.filter.FilterFactory;
import org.codehaus.activemq.filter.FilterFactoryImpl;
import org.codehaus.activemq.filter.NoLocalFilter;
import org.codehaus.activemq.message.ActiveMQMessage;
import org.codehaus.activemq.message.ConsumerInfo;
import org.codehaus.activemq.message.MessageAck;
import org.codehaus.activemq.message.util.MemoryBoundedQueue;
import org.codehaus.activemq.message.util.MemoryBoundedQueueManager;
import org.codehaus.activemq.service.MessageContainer;
import org.codehaus.activemq.service.MessageContainerManager;

import javax.jms.JMSException;
import java.util.Iterator;

/**
 * A MessageContainerManager for transient topics
 * 
 * @version $Revision: 1.6 $
 */

/**
 * A manager of MessageContainer instances
 */
public class TransientTopicBoundedMessageManager implements MessageContainerManager {
    private MemoryBoundedQueueManager queueManager;
    private ConcurrentHashMap containers;
    private FilterFactory filterFactory;
    private SynchronizedBoolean started;

    /**
     * Constructor for TransientTopicBoundedMessageManager
     *
     * @param mgr
     */
    public TransientTopicBoundedMessageManager(MemoryBoundedQueueManager mgr) {
        this.queueManager = mgr;
        this.containers = new ConcurrentHashMap();
        this.filterFactory = new FilterFactoryImpl();
        this.started = new SynchronizedBoolean(false);
    }

    /**
     * start the manager
     *
     * @throws JMSException
     */
    public void start() throws JMSException {
        if (started.commit(false, true)) {
            for (Iterator i = containers.values().iterator(); i.hasNext();) {
                TransientTopicBoundedMessageContainer container = (TransientTopicBoundedMessageContainer) i.next();
                container.start();
            }
        }
    }

    /**
     * stop the manager
     *
     * @throws JMSException
     */
    public void stop() throws JMSException {
        if (started.commit(true, false)) {
            for (Iterator i = containers.values().iterator(); i.hasNext();) {
                TransientTopicBoundedMessageContainer container = (TransientTopicBoundedMessageContainer) i.next();
                container.stop();
            }
        }
    }

    /**
     * Add a consumer if appropiate
     *
     * @param client
     * @param info
     * @throws JMSException
     */
    public void addMessageConsumer(BrokerClient client, ConsumerInfo info) throws JMSException {
        if (info.getDestination().isTopic()) {
            TransientTopicBoundedMessageContainer container = (TransientTopicBoundedMessageContainer) containers
                    .get(client);
            if (container == null) {
                MemoryBoundedQueue queue = queueManager.getMemoryBoundedQueue(client.toString());
                container = new TransientTopicBoundedMessageContainer(client, queue);
                containers.put(client, container);
                if (started.get()) {
                    container.start();
                }
            }
            container.addConsumer(createFilter(info), info);
        }
    }

    /**
     * @param client
     * @param info
     * @throws JMSException
     */
    public void removeMessageConsumer(BrokerClient client, ConsumerInfo info) throws JMSException {
        if (info.getDestination().isTopic()) {
            TransientTopicBoundedMessageContainer container = (TransientTopicBoundedMessageContainer) containers
                    .get(client);
            if (container != null) {
                container.removeConsumer(info);
                if (container.isInactive()) {
                    containers.remove(client);
                    container.close();
                }
            }
        }
    }

    /**
     * Delete a durable subscriber
     *
     * @param clientId
     * @param subscriberName
     * @throws JMSException if the subscriber doesn't exist or is still active
     */
    public void deleteSubscription(String clientId, String subscriberName) throws JMSException {
    }

    /**
     * @param client
     * @param message
     * @throws JMSException
     */
    public void sendMessage(BrokerClient client, ActiveMQMessage message) throws JMSException {
        if (message != null && message.getJMSActiveMQDestination().isTopic()) {
            for (Iterator i = containers.values().iterator(); i.hasNext();) {
                TransientTopicBoundedMessageContainer container = (TransientTopicBoundedMessageContainer) i.next();
                container.targetAndDispatch(message.shallowCopy());
                //System.out.println("ENQ msg(" + message.getMemoryUsage() + ") total cap = " + queueManager.getCurrentCapacity());
            }
        }
    }

    public void acknowledgeMessage(BrokerClient client, MessageAck ack) throws JMSException {
    }

    public void acknowledgeTransactedMessage(BrokerClient client, String transactionId, MessageAck ack) throws JMSException {
    }

    public void redeliverMessage(BrokerClient client, MessageAck ack) throws JMSException {
    }

    public void poll() throws JMSException {
    }

    /**
     * A hook when the transaction is about to be commited; so apply all outstanding commands to the Journal if using a
     * Journal (transaction log)
     *
     * @param client
     * @param transactionId
     * @throws JMSException
     */
    public void commitTransaction(BrokerClient client, String transactionId) throws JMSException {
    }

    /**
     * A hook when the transaction is about to be rolled back; so discard all outstanding commands that are pending to
     * be written to the Journal
     *
     * @param client
     * @param transactionId
     */
    public void rollbackTransaction(BrokerClient client, String transactionId) {
    }

    public MessageContainer getContainer(String physicalName) {
        return null;
    }

    /**
     * Create filter for a Consumer
     *
     * @param info
     * @return the Fitler
     * @throws javax.jms.JMSException
     */
    protected Filter createFilter(ConsumerInfo info) throws JMSException {
        Filter filter = filterFactory.createFilter(info.getDestination(), info.getSelector());
        if (info.isNoLocal()) {
            filter = new AndFilter(filter, new NoLocalFilter(info.getClientId()));
        }
        return filter;
    }
}