/** 
 * 
 * Copyright 2004 Protique Ltd
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License. 
 * 
 **/

package org.codehaus.activemq.transport;

import EDU.oswego.cs.dl.util.concurrent.CopyOnWriteArrayList;
import EDU.oswego.cs.dl.util.concurrent.Executor;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.codehaus.activemq.message.Packet;
import org.codehaus.activemq.message.PacketListener;
import org.codehaus.activemq.message.Receipt;
import org.codehaus.activemq.message.ReceiptHolder;
import org.codehaus.activemq.util.ExecutorHelper;

import javax.jms.ExceptionListener;
import javax.jms.JMSException;
import java.net.URI;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * Some basic functionality, common across most transport implementations of channels
 *
 * @version $Revision: 1.10 $
 */
public abstract class AbstractTransportChannel implements TransportChannel {
    private static final Log log = LogFactory.getLog(AbstractTransportChannel.class);

    private CopyOnWriteArrayList listeners = new CopyOnWriteArrayList();
    private HashMap requestMap = new HashMap();
    private PacketListener packetListener;
    private ExceptionListener exceptionListener;
    private String clientID;
    private TransportChannelListener transportChannelListener;


    /**
     * close the channel
     */
    public void stop() {
        Map map = (Map) this.requestMap.clone();
        for (Iterator i = map.values().iterator(); i.hasNext();) {
            ReceiptHolder rh = (ReceiptHolder) i.next();
            rh.close();
        }
        map.clear();
        requestMap.clear();
        if (transportChannelListener != null) {
            transportChannelListener.removeClient(this);
        }
        exceptionListener = null;
        packetListener = null;
    }

    /**
     * synchronously send a Packet
     *
     * @param packet
     * @return a Receipt
     * @throws JMSException
     */
    public Receipt send(Packet packet) throws JMSException {
        return send(packet, 0);
    }

    /**
     * Synchronously send a Packet
     *
     * @param packet  packet to send
     * @param timeout amount of time to wait for a receipt
     * @return the Receipt
     * @throws JMSException
     */
    public Receipt send(Packet packet, int timeout) throws JMSException {
        ReceiptHolder rh = new ReceiptHolder();
        requestMap.put(packet.getId(), rh);
        doAsyncSend(packet);
        Receipt result = rh.getReceipt(timeout);
        return result;
    }


    // Properties
    //-------------------------------------------------------------------------

    public TransportChannelListener getTransportChannelListener() {
        return transportChannelListener;
    }

    public void setTransportChannelListener(TransportChannelListener transportChannelListener) {
        this.transportChannelListener = transportChannelListener;
    }

    /**
     * Add a listener for changes in a channels status
     *
     * @param listener
     */
    public void addTransportStatusEventListener(TransportStatusEventListener listener) {
        listeners.add(listener);
    }

    /**
     * Remove a listener for changes in a channels status
     *
     * @param listener
     */
    public void removeTransportStatusEventListener(TransportStatusEventListener listener) {
        listeners.remove(listener);
    }

    public String getClientID() {
        return clientID;
    }

    public void setClientID(String clientID) {
        this.clientID = clientID;
    }

    public ExceptionListener getExceptionListener() {
        return exceptionListener;
    }

    public PacketListener getPacketListener() {
        return packetListener;
    }

    /**
     * Set a listener for Packets
     *
     * @param l
     */
    public void setPacketListener(PacketListener l) {
        this.packetListener = l;
    }


    /**
     * Set an exception listener to listen for asynchronously generated exceptions
     *
     * @param listener
     */
    public void setExceptionListener(ExceptionListener listener) {
        this.exceptionListener = listener;
    }

    // Implementation methods
    //-------------------------------------------------------------------------

    /**
     * consume a packet from the channel
     *
     * @param packet
     */
    protected void doConsumePacket(Packet packet) {
        if (!doHandleReceipt(packet)) {
            if (packetListener != null) {
                packetListener.consume(packet);
            }
            else {
                log.warn("No packet listener set to receive packets");
            }
        }
    }

    protected boolean doHandleReceipt(Packet packet) {
        boolean result = false;
        if (packet != null) {
            if (packet.isReceipt()) {
                result = true;
                Receipt receipt = (Receipt) packet;
                ReceiptHolder rh = (ReceiptHolder) requestMap.remove(receipt.getCorrelationId());
                if (rh != null) {
                    rh.setReceipt(receipt);
                }
                else {
                    log.warn("No Packet found to match Receipt correlationId: " + receipt.getCorrelationId());
                }
            }
        }
        return result;
    }

    /**
     * send a Packet to the raw underlying transport
     * This method is here to allow specific implementations
     * to override this method
     *
     * @param packet
     * @throws JMSException
     */
    protected void doAsyncSend(Packet packet) throws JMSException {
        asyncSend(packet);
    }

    /**
     * Handles an exception thrown while performing async dispatch of messages
     *
     * @param e
     */
    protected void onAsyncException(JMSException e) {
        if (exceptionListener != null) {
            exceptionListener.onException(e);
        }
        else {
            log.warn("Caught exception dispatching message and no ExceptionListener registered: " + e, e);
        }
    }

    /**
     * Fire status event to any status event listeners
     *
     * @param remoteURI
     * @param status
     */
    protected void fireStatusEvent(URI remoteURI, int status) {
        TransportStatusEvent event = new TransportStatusEvent();
        event.setChannelStatus(status);
        event.setRemoteURI(remoteURI);
        fireStatusEvent(event);
    }

    /**
     * Fire status event to any status event listeners
     *
     * @param event
     */
    protected void fireStatusEvent(TransportStatusEvent event) {
        if (event != null) {
            for (Iterator i = listeners.iterator(); i.hasNext();) {
                TransportStatusEventListener l = (TransportStatusEventListener) i.next();
                l.statusChanged(event);
            }
        }
    }

    /**
     * A helper method to stop the execution of an executor
     *
     * @param executor the executor or null if one is not created yet
     */
    protected void stopExecutor(Executor executor) throws InterruptedException, JMSException {
        ExecutorHelper.stopExecutor(executor);
    }
}
